#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#      MCcoordstore - simple web based store for Minecraft points of interest
#      Copyright (C) 2021  Daniel Philip Weatherill

#      This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.

#      This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.

#      You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



"""
Created on Thu Dec  9 17:55:36 2021

@author: danw
"""

from flask.cli import with_appcontext
from flask_sqlalchemy import SQLAlchemy
from datetime import datetime, timezone
import click
from typing import Sequence
from werkzeug.security import generate_password_hash
from flask_login import UserMixin

db = SQLAlchemy()

class User(UserMixin, db.Model):
    __tablename__ = "user"
    userid = db.Column(db.Integer, primary_key = True)
    username = db.Column(db.String(80), unique=True, nullable=False)
    displayname = db.Column(db.String(80), unique=True, nullable=False)
    hashed_pw = db.Column(db.String, nullable=False)
    is_admin = db.Column(db.Boolean, default=False)
    
    @classmethod
    def create_new_user(cls, username: str, displayname: str, password: str) -> "User":
        pwhash = generate_password_hash(password)
        return User(username = username, displayname=displayname, hashed_pw=pwhash)
    
    def get_id(self):
        return str(self.userid)



class RenderStyle(db.Model):
    __tablename__ = "renderstyle"
    styleid = db.Column(db.Integer, primary_key = True)
    name = db.Column(db.String(80), nullable=True)
    style = db.Column(db.PickleType)

class PointOfInterest(db.Model):
    __tablename__ = "pointofinterest"
    poiid = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(80), nullable=False)
    create_date = db.Column(db.DateTime(timezone=True), nullable=False, default=lambda : datetime.now(timezone.utc))
    public = db.Column(db.Boolean, default=False)

    userid = db.Column(db.Integer, db.ForeignKey("user.userid"), nullable=False)
    user = db.relationship("User", backref=db.backref("pois"))
    style = db.Column(db.Integer, db.ForeignKey("renderstyle.styleid"))

    coord_x = db.Column(db.Integer)
    coord_y = db.Column(db.Integer)
    coord_z = db.Column(db.Integer)
    
    @property
    def coords(self):
        return (self.coord_x, self.coord_y, self.coord_z)
    
    @coords.setter
    def coords(self, val: Sequence[int]):
        if len(val) != 3:
            raise ValueError("invalid length of coordinate")
        self.coord_x = val[0]
        self.coord_y = val[1]
        self.coord_z = val[2]
        
    def serialize_to_dict(self):
        return {"id" : self.poiid,
                "name" : self.name,
                "created" : str(self.create_date),
                "public" : self.public,
                "username" : self.user.username,
                "coord" : self.coords
                }


def get_db(app):
    db.init_app(app)
    return db


def create_db(admin_pass, db):
    db_path = db.engine.url
    print(f"creating new database with path: {db_path}")
    db.create_all()

    print("creating default admin and guest users")
    admin_user = User.create_new_user("admin", "admin", admin_pass)
    admin_user.is_admin = True
    guest_user = User.create_new_user("guest", "guest", "guest")

    db.session.add(admin_user)
    db.session.add(guest_user)
    db.session.commit()
    


@click.command("create-db")
@click.argument("admin_pass")
@with_appcontext
def create_db_command(admin_pass):
    from flask import current_app
    db.init_app(current_app)
    create_db(admin_pass, db)
    
    


@click.command("change-pw")
@click.argument("username")
@click.argument("new_pass")
@with_appcontext
def change_pw_command(username, new_pass):
    from flask import current_app
    db.init_app(current_app)

    usr = User.query.filter_by(username=username).first()
    pwhash = generate_password_hash(new_pass)
    usr.hashed_pw = pwhash
    db.session.commit()

    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#      MCcoordstore - simple web based store for Minecraft points of interest
#      Copyright (C) 2021  Daniel Philip Weatherill

#      This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.

#      This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.

#      You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



"""
Created on Thu Dec  9 18:23:28 2021

@author: danw
"""

from flask import render_template, redirect, flash, current_app, g
from markupsafe import Markup
from sqlalchemy import select
from werkzeug.security import check_password_hash
from flask_login import login_required, current_user, login_user, logout_user

from MCcoordstore import create_app
from .plots import LocationsPlot
from .forms import AddPOIForm, SignupForm, LoginForm
from .db import get_db, User, PointOfInterest
from .utils import serialize_pois

app = create_app()
db = get_db(app)

@app.route("/", methods=["POST", "GET"])
def index():
    locplot = LocationsPlot()


    if current_user.is_authenticated:
        pois = PointOfInterest.query.all()
    else:
        pois = PointOfInterest.query.filter_by(public=True).all()
    

    locplot.xdat = [_.coords[0] for _ in pois]
    locplot.zdat = [_.coords[2] for _ in pois]
    locplot.update_plot()

    form = AddPOIForm()
    if form.validate_on_submit():
        #TODO: user guest only for now
        poi = PointOfInterest(name=form.data["name"], public=form.data["public"], user=current_user)
        poi.coords = form.coords
        db.session.add(poi)
        db.session.commit()
        return redirect("/")

    return render_template("index.htm", map_html_2d = Markup(locplot.rendered_html),
                           poi_table_data = pois, form=form)


@app.route("/add_manual", methods=["POST","GET"])
@login_required
def add_manual():
    form = AddPOIForm()
    if form.validate_on_submit():
        #TODO: user guest only for now
        poi = PointOfInterest(name=form.data["name"], public=form.data["public"], user=current_user)
        poi.coords = form.coords
        db.session.add(poi)
        db.session.commit()
        return redirect("/")

    return render_template("add_manual.htm", form=form)

@app.route("/signup", methods=["POST","GET"])
def signup():
    form = SignupForm()
    if form.validate_on_submit():
        #check user is unique
        existing_usrname = User.query.filter_by(username=form.data["username"]).first()
        if existing_usrname:
            flash("either username or displayname is already in use!", "flash-error")
            return redirect("/signup")

        existing_displayname = User.query.filter_by(displayname=form.data["displayname"]).first()
        if existing_displayname:
            flash("displayname is already in use!", "flash-error")
            return redirect("/signup")

        new_usr = User.create_new_user(form.data["username"], form.data["displayname"],
                                       form.data["password"])
        db.session.add(new_usr)
        db.session.commit()

        flash("signup successful, you can now login", "flash-success")
        return redirect("/")
    return render_template("signup.htm", form=form)

@app.route("/login", methods=["POST","GET"])
def login():
    form = LoginForm()
    if form.validate_on_submit():
        user = User.query.filter_by(username=form.data["username"]).first()
        if not user or not check_password_hash(user.hashed_pw, form.data["password"]):
            flash("unable to login!", "flash-error")
            return redirect("/login")
        
        login_user(user)
        flash("successfully logged in", "flash-success")
        return redirect("/")
    return render_template("login.htm", form=form)


#TODO: implement REST DELETE METHOD
@app.route("/poi/delete/<idx>", methods=["GET"])
@login_required
def delete_poi(idx):
    try:
        intid = int(idx)
    except ValueError as err:
        flash("unable to interpret idx as int", "flash-error")
        return redirect("/")
    
    poi = PointOfInterest.query.filter_by(poiid=intid).first()
    
    if not poi:
        flash("invalid POI idx", "flash-error")
        return redirect("/")
    
    if poi.user != current_user and not current_user.is_admin:
        flash("no permission to delete POI!", "flash-error")
        return redirect("/")

    db.session.delete(poi)
    db.session.commit()

    flash("POI deleted", "flash-success")
    return redirect("/")



@app.route("/logout")
@login_required
def logout():
    logout_user()
    return redirect("/")

@app.route("/dump_pois")
@login_required
def dump_pois():
    return serialize_pois(app, db)

    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#      MCcoordstore - simple web based store for Minecraft points of interest
#      Copyright (C) 2021  Daniel Philip Weatherill

#      This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.

#      This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.

#      You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



"""
Created on Fri Dec 10 21:28:26 2021

@author: danw
"""

from .db import PointOfInterest
import json

def serialize_pois(app, db):
    pois = PointOfInterest.query.all()
    poilist = [_.serialize_to_dict() for _ in pois]
    return json.dumps(poilist)



    
    
    
    
    



from flask import Flask, g
import os
from typing import Dict, Any
from .db import create_db_command, User, change_pw_command, get_db, create_db
from flask_login import LoginManager
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy

CONFIG_ENVVAR = "MCCOORDSTORE_CONFIG"

def create_app(test_config: Dict[str, Any]=None):
    app = Flask(__name__, instance_relative_config=True)
    
    if test_config is None:
        if os.environ.get(CONFIG_ENVVAR) is None:
            print("WARNING: using default test config! DO NOT USE IN PRODUCTION!")
            app.config.from_object("MCcoordstore.default_config.DefaultConfig")
        else:
            app.config.from_envvar(CONFIG_ENVVAR, silent=False)
    else:
        app.config.from_mapping(test_config)

    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass
    
    app.cli.add_command(create_db_command)
    app.cli.add_command(change_pw_command)
    
    
    lman = LoginManager()
    lman.login_view = "login"
    lman.init_app(app)
    
    @lman.user_loader
    def load_user(user_id: str):
        return User.query.get(int(user_id))
        
    db = get_db(app)

    
    migrate = Migrate(app,db)
    
    return app

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#      MCcoordstore - simple web based store for Minecraft points of interest
#      Copyright (C) 2021  Daniel Philip Weatherill

#      This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.

#      This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.

#      You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Created on Thu Dec  9 19:05:14 2021

@author: danw
"""

import matplotlib
import matplotlib.pyplot as plt
import mpld3
from typing import List, Union
import numpy as np

matplotlib.use("template")


class LocationsPlot:
    def __init__(self, **plotkwargs):
        self.fig, self.ax = plt.subplots(1,1, **plotkwargs)
        self.scatter = None
        self.xdat = []
        self.zdat = []
        self.syms = []
        self.colors = []
        self._needs_render: Union[bool,None] = None
        self._rendered_html = None

    def update_plot(self):
        if self.scatter is None:
            self.scatter = self.ax.scatter(self.xdat, self.zdat)
        else:
            dat = np.vstack((self.xdat, self.zdat))
            self.scatter.set_offsets(dat)
        
        self._needs_render = True
        
    @property
    def rendered_html(self) -> str:
        if self._needs_render is None:
            raise RuntimeError("plot needs to be updated first")
        if self._needs_render:
            self._rendered_html = mpld3.fig_to_html(self.fig)
            self._needs_render = False
        return self._rendered_html

        
    
        
if __name__ == "__main__":
    plt.close("all")
    newplot = LocationsPlot()
    
    newplot.xdat = np.random.normal(0., 1., 5)
    newplot.zdat = np.random.normal(0, 1, 5)
    
    newplot.update_plot()
    

    



#      MCcoordstore - simple web based store for Minecraft points of interest
#      Copyright (C) 2021  Daniel Philip Weatherill

#      This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.

#      This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.

#      You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from setuptools import setup, find_packages
import pathlib

HERE = pathlib.Path(__file__).parent
README = (HERE / "README").read_text()

setup(name="MCcoordstore",
      version="0.0.1dev",
      description="simple coordinate store for minecraft",
      long_description=README,
      long_description_content_type="text/markdown",
      license="AGPL-3.0-or-later",
      author="Dan Weatherill",
      author_email="dan.weatherill@cantab.net",
      url="https://github.com/weatherhead99/MCcoordstore",
      packages = find_packages(),
      include_package_data = True,
      classifiers = [
          "Development Status :: 2 - Pre-Alpha",
          "Framework :: Flask",
          "License :: OSI Approved :: GNU Affero General Public License v3",
          "Topic :: Internet :: WWW/HTTP :: WSGI :: Application"
          ],
      install_requires = ["flask",
                          "flask_sqlalchemy",
                          "Flask-WTF",
                          "Flask-login",
                          "matplotlib",
                          "numpy",
                          "mpld3",
                          "werkzeug",
                          "flask-migrate"]
      )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#      MCcoordstore - simple web based store for Minecraft points of interest
#      Copyright (C) 2021  Daniel Philip Weatherill

#      This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your option)
# any later version.

#      This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.

#      You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Created on Thu Dec  9 22:33:46 2021

@author: danw
"""

from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, PasswordField, BooleanField
from wtforms.validators import DataRequired, EqualTo

def _reqfield(func, *args, **kwargs):
    return func(*args, **kwargs, validators=[DataRequired()])

class AddPOIForm(FlaskForm):
    name = _reqfield(StringField, "name")
    x = _reqfield(IntegerField, "x")
    y = _reqfield(IntegerField, "y")
    z = _reqfield(IntegerField, "z")
    public = BooleanField("public")
    
    @property
    def coords(self):
        return (self.data["x"], self.data["y"], self.data["z"])
    

class SignupForm(FlaskForm):
    username = _reqfield(StringField, "username")
    displayname = _reqfield(StringField, "display name")
    password = PasswordField("password", [DataRequired(), 
                                          EqualTo("confirm", message="passwords must match")])
    confirm  = PasswordField("confirm password", [DataRequired()])

class LoginForm(FlaskForm):
    username = _reqfield(StringField, "username")
    password = _reqfield(PasswordField, "password")
    remember = BooleanField("remember me")

    
